#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mach.h>
#include <dynld.h>

#define OP_PREFIX	  "op_"
#define OP_PREFIX_SIZE 3

#define SPECIALOP_PUSH       "__op_p9_push"
#define SPECIALOP_STORESTACK "__op_p9_storestack"

#define CONSTFOLD "__op_param"
#define JUMPLABEL "__op_gen_label"

#define MAX_CFOLDS	3

static void
fatal(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprint(2, fmt, ap);
	va_end(ap);
	exits("error");
}

static void
nullprint(char *fmt, ...)
{
	USED(fmt);
}

typedef struct Rrec Rrec;
struct Rrec {
	Dynrel *reloc;
	Rrec *next;
};

typedef struct Funct Funct;
struct Funct {
	Sym *sym;
	uvlong bounds[2];
	char *qemu_op_name;
	uvlong size;
	int maxcfold;
	Rrec *rrecs;
	Funct *next;
};

typedef struct DynsymMetadata DynsymMetadata;
struct DynsymMetadata {
	int cfold_index;	/* -1 for none */
	int label_index;	/* -1 for none */
};

int infd;
Biobuf outbio;
Fhdr fh;

uchar *aout_text;

ulong nimport;
Dynsym *import_table;						/* As given by dynld */
DynsymMetadata *import_table_meta;			/* Same index as import_table */

ulong nreloc;
Dynrel *reloc_table;

Funct *microops;				/* Sorted by address */

Funct *specialop_storestack;	/* Ignore chain */
Funct *specialop_push;			/* Ignore chain */

//Rrec *homeless_rrecs = nil;

int paranoia;
int verbosity;

void (*paranoid)(char *, ...) = fatal;

enum {
	MODE_USAGE,				/* Help */
	MODE_OP_DEFINES,		/* Emit DEF() declarations */
	MODE_OP_GENERATORS,		/* Emit gen_...() functions */
	MODE_TRANSLATOR,		/* Emit the translator */
	MODE_TABLES,			/* Emit import tables */
	MODE_OP_BODIES,			/* Emit arrays for micro-ops */
} mode;

static void
verbose(char *fmt, ...)
{
	va_list ap;

	if(verbosity == 0)
		return;

	va_start(ap, fmt);
	Bvprint(&outbio, fmt, ap);
	va_end(ap);
}

static void
emit(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	Bvprint(&outbio, fmt, ap);
	va_end(ap);
}

static void
load_text(void)
{
	/* Load file data */
	seek(infd, fh.txtoff, 0);
	aout_text = malloc(fh.txtsz);
	if(!aout_text)
		fatal("Out of memory: text region");
	if(read(infd, aout_text, fh.txtsz) != fh.txtsz)
		fatal("Unable to read text region");
}

void
insert_function(Funct **head, Funct *new)
{
	if(*head == nil)
   		*head = new;
   	else {
		Funct *fi = *head, *fp = nil;
		while(fi && fi->sym->value < new->sym->value){
			fp = fi;
   			fi = fi->next;
   		}
 
		if (!fp){
   			new->next = *head;
			*head = new;
		} else {
   			new->next = fp->next;
   			fp->next = new;
   		}
	}
}

void
filter_symbols(void)
{
	long n;

	Sym* s = symbase(&n);

	while(n--){
		verbose("/* Symbol %c %s @ 0x%.8llx */\n",
			s->type, s->name, s->value);

		uvlong bounds[2];

		switch(s->type)
		{
		case 'T':	/* Exported text */
		case 't':	/* Static text */
			if(fnbound(s->value, bounds) == 1){
				/* Smells like a function */

				verbose("/*  Is function: [0x%.8ullx, 0x%.8ullx] = 0x%ullx */\n",
					bounds[0], bounds[1], bounds[1] - bounds[0]);
	
   				Funct *fn = malloc(sizeof(Funct));
				if(!fn)
					fatal("Out of memory allocating a function");
	
   				fn->sym = s;
				fn->bounds[0] = bounds[0];
				fn->bounds[1] = bounds[1];
				fn->size = fn->bounds[1] - fn->bounds[0];
				fn->maxcfold = 0;
   				fn->rrecs = nil;
				fn->next = nil;

				if(s->type == 't')
					paranoid("// Static function: %s\n", s->name);

				if(strstr(s->name, OP_PREFIX) == s->name){
					fn->qemu_op_name = s->name+OP_PREFIX_SIZE;
				} else {
					fn->qemu_op_name = nil;

					if(!strcmp(s->name, SPECIALOP_STORESTACK))
						specialop_storestack = fn;
					else if(!strcmp(s->name, SPECIALOP_PUSH))
						specialop_push = fn;
				}

				insert_function(&microops, fn);
			}
			break;
		case 'd':
			paranoid("/* Static data: %s */\n", s->name);
			break;
		case 'D':
			if(!strcmp(s->name, ".string")){
				verbose("/* Exported string table.  I cannot tell if this is OK. */\n");
			}else if(!strcmp(s->name, "_exporttab")){
				verbose("/* DLM export table */\n");
			}else{
				//paranoid("// Exported data: %s\n", s->name);
				verbose("/* Exported data: %s */\n", s->name);
			}
			break;
		case 'b':
			paranoid("/* Static bss: %s */\n", s->name);
			break;
		case 'B':
			if(!strcmp(s->name, "end")){
				verbose("/* Exported end marker */\n");
			}else if(!strcmp(s->name, "edata")){
				verbose("/* Exported end of data marker */\n");
			}else{
				paranoid("/* Exported bss: %s */\n", s->name);
			}
		default:
			break;
		}
		s++;
	}
}

void
emit_byte_array(char *name, uchar* start, vlong size)
{
	emit("char %s[] = { \n", name);
	switch((int)(size % 8))
	{
	case 0:	do {
	case 7:		emit(" 0x%.2x, ", *start++);
	case 6:		emit(" 0x%.2x, ", *start++);
	case 5:		emit(" 0x%.2x, ", *start++);
	case 4:		emit(" 0x%.2x, ", *start++);
	case 3:		emit(" 0x%.2x, ", *start++);
	case 2:		emit(" 0x%.2x, ", *start++);
	case 1:		emit(" 0x%.2x,\n", *start++);
			} while((size-=8) > 0);
	}
	emit("};\n");
}

void 
emit_functions(void)
{
	Funct *fi;
	for(fi = microops; fi != nil; fi = fi->next){

		/* Don't emit things that aren't micro-ops */
		if(fi->qemu_op_name == nil)
			continue;

		emit_byte_array(fi->sym->name,
			aout_text + fi->bounds[0] - fh.txtaddr,
			fi->bounds[1] - fi->bounds[0]);
	}

	fi=specialop_push;
	if(fi)
		emit_byte_array(fi->sym->name,
			aout_text + fi->bounds[0] - fh.txtaddr,
			fi->bounds[1] - fi->bounds[0]);

	fi=specialop_storestack;
	if(fi)
		emit_byte_array(fi->sym->name,
			aout_text + fi->bounds[0] - fh.txtaddr,
			fi->bounds[1] - fi->bounds[0]);
}

void
create_import_meta(void)
{
	ulong n;

	import_table_meta = malloc(nimport * sizeof(DynsymMetadata));
	if(import_table_meta == nil){
		fatal("Out of memory for import metadata\n");
	}

	for(n = 0; n < nimport; n++){
		if(strstr(import_table[n].name, CONSTFOLD) == import_table[n].name){
			int which = atoi( import_table[n].name + strlen(CONSTFOLD) );
			verbose("/* Found constant folder %d index %uld */\n", which, n);
			import_table_meta[n].cfold_index = which;
		} else {
			import_table_meta[n].cfold_index = -1;
		}

		if(strstr(import_table[n].name, JUMPLABEL) == import_table[n].name){
			int which = atoi( import_table[n].name + strlen(JUMPLABEL) );
			verbose("/* Found jump label %d index %uld */\n", which, n);
			import_table_meta[n].label_index = which;
		} else {
			import_table_meta[n].label_index = -1;
		}
	}
}

void
load_imports(void)
{
	seek(infd, 0, 0);
	nimport = dyn_import_table(infd, &fh, &import_table);
}

void
load_relocs(void)
{
	seek(infd, 0, 0);
	nreloc = dyn_reloc_table(infd, &fh, &reloc_table);
}

void
pair_relocs(void)
{
	ulong n;
	Funct *fi;
	Rrec *ri;
	Dynrel *cdr;

	for(n = 0, cdr = reloc_table; n < nreloc; n++, cdr++)
	{
		verbose("/* Reloc %ld @ 0x%ulx mode"
			" %d, idx %d off 0x%llx */\n",
			n, cdr->offset, cdr->mode,
			cdr->reloc_index, cdr->reloc_offset);


		for(fi = microops;
			fi != nil && cdr->offset > fi->bounds[1];
			fi = fi->next)
		{ ; }

		if(!fi)
		{
			verbose("  /* Is not associated with a function */\n");
		}else{
			int ii = cdr->reloc_index;
			if(ii == -1){
				verbose("  /* Fun %s %s [NO SYMBOL] */\n",
					fi->sym->name,
					((fi->sym->type == 't') ? "[WARNING: STATIC]" : ""));
			} else {
				int cfi = import_table_meta[ii].cfold_index;
				int li = import_table_meta[ii].label_index;

				verbose("  /* Fun %s%s %d %d %d */ \n",
					fi->sym->name,
					((fi->sym->type == 't') ? "[WARNING: STATIC]" : ""),
					ii, cfi, li);

				if(cfi > fi->maxcfold)
					fi->maxcfold = cfi;

				if(li > fi->maxcfold)
					fi->maxcfold = li;

			}

			if(cdr->mode == 0)
				paranoid(
					"/* I feel nervous (but can't really know) about a mode 0 relocation */\n"
				);
	
			ri = malloc(sizeof(Rrec));
			ri->reloc = cdr;
			ri->next = fi->rrecs;
			fi->rrecs = ri;
		}



	}
}

void
check_special_op(Funct *fi)
{
	if(fi->maxcfold != 1)
		fatal("Special op %s should take exactly one constant fold"
				" (not %d)\n",
				fi->sym->name,
				fi->maxcfold);
	
	if(fi->rrecs == nil)
		fatal("Special op %s has no relocation records\n",
			fi->sym->name);

	if(fi->rrecs->next != nil)
		fatal("Special op %s has more than one relocation record\n",
			fi->sym->name);

	if(import_table_meta[fi->rrecs->reloc->reloc_index].label_index != -1)
		fatal("Special op %s has a label-style relocation\n",
			fi->sym->name);

	if(import_table_meta[fi->rrecs->reloc->reloc_index].cfold_index == -1)
		fatal("Special op %s does not have a cfold-style relocation\n",
			fi->sym->name);
}

void
check_special_ops(void)
{
	if(specialop_push == nil)
		fatal("Did not find a special push operation.\n");

	check_special_op(specialop_push);

	if(specialop_storestack == nil)
		fatal("Did not find a special stack storing operation.\n");

	check_special_op(specialop_storestack);
}

void
check_cfolds(void)
{
	Funct *fi;
	Rrec *ri;
	int search;

	verbose("/* Contant folder check ... */\n");
	
	for(fi = microops; fi; fi=fi->next){
		if(fi->maxcfold == 0)
			continue;
		if(fi->qemu_op_name == nil)
			continue;

		if(fi->maxcfold > MAX_CFOLDS)
			fatal("Too many parameters in %s: %d\n", fi->sym->name, fi->maxcfold);
		
		for(search = fi->maxcfold; search > 0; search--){
			/* First, is it a constant fold? */
			for(ri = fi->rrecs; ri; ri = ri->next){
				if(import_table_meta[ri->reloc->reloc_index].cfold_index == search)
					break;
			}
			if(!ri){
				/* Nope... is it a label? */
				for(ri = fi->rrecs; ri; ri = ri->next){
					if(import_table_meta[ri->reloc->reloc_index].label_index == search)
						break;
				}
				if(!ri)
					fatal("Out of order parameters in %s: %d\n", fi->sym->name, search);
			}else{
				/* Yep... is it also a label? */
				for(ri = fi->rrecs; ri; ri = ri->next){
					if(import_table_meta[ri->reloc->reloc_index].label_index == search)
						break;
				}
				if(ri)
					fatal("Duplicate label and cfold in %s: %d\n", fi->sym->name, search);
			}
		}
	}

	verbose("/* ... done (looks good). */\n");
}

void
emit_import_table(void)
{
	ulong n;

	/* Write out emulation of the import table */
	emit("#include \"plan9/includeonce.h\"\n");
	emit("#include \"plan9/dyngen-helpers-plan9.h\"\n");
	emit("int dyngen_nimport = %uld;\n", nimport);
	emit("struct DyngenDynsym dyngen_itab_elts[] = {\n");
	for(n = 0; n < nimport; n++){
		int abuse = 0;
		if((import_table_meta[n].cfold_index != -1)
			|| (import_table_meta[n].label_index != -1))
				abuse = 1;

			emit("\t{.d = { .sig = 0x%.8ulx, .addr = nil, .name = \"%s\" }, .abuse = %d},\n",
						import_table[n].sig,
						import_table[n].name,
						abuse);
	}
	emit("};\n");

	emit("Dynsym *dyngen_itab[] = {\n");
	for(n = 0; n < nimport; n++){
		emit("\t&(dyngen_itab_elts[%uld].d),\n", n);
	}
	emit("};\n");
}

void
verbose_import_table(void)
{
	ulong n;

	/* Write out some english about the import table */
	emit("/* Found %uld imports */\n", nimport);
	for(n = 0; n < nimport; n++){
		if(import_table_meta[n].cfold_index != -1)
			emit("/* Index %uld: %s [cfolder %d] */\n",
						n,
						import_table[n].name,
						import_table_meta[n].cfold_index
			);
		else if(import_table_meta[n].label_index != -1)
			emit("/* Index %uld: %s [label %d] */\n",
						n,
						import_table[n].name,
						import_table_meta[n].label_index
			);
		else
			emit("/* Index %uld: %s */\n",
						n,
						import_table[n].name);
	}
}

void
emit_definitions(uvlong padding)
{
	Funct *fi;

	/* Some foolishness we're expected to emit */
	emit("DEF(end, 0, 0)\n");
	emit("DEF(nop, 0, 0)\n");
	emit("DEF(nop1, 1, 0)\n");
	emit("DEF(nop2, 2, 0)\n");
	emit("DEF(nop3, 3, 0)\n");

	/* Plan 9 special operations */
	emit("DEF(p9_storestack, 1, %d)\n", specialop_storestack->size);

	for(fi = microops; fi; fi=fi->next)
	{
		if(fi->qemu_op_name == nil)
			continue;

		emit("DEF(%s, %d, %ulld)\n",
			fi->qemu_op_name,
			fi->maxcfold,
			fi->size + padding
		);
	}
}

void
emit_generators(void)
{
	Funct *fi;
	int i;

	for(fi = microops; fi; fi=fi->next)
	{
		/* Don't emit generators for things that aren't micro-ops */
		if(fi->qemu_op_name == nil)
			continue;

		emit("static inline void gen_%s(", fi->sym->name);
		if (fi->maxcfold == 0) {
			emit("void");
		} else {
			for(i = 0; i < fi->maxcfold; i++) {
				if (i != 0)
					emit(", ");
				emit("long param%d", i + 1);
			}
		}
		emit(")\n");
		emit("{\n");
		for(i = 0; i < fi->maxcfold; i++) {
			emit("    *gen_opparam_ptr++ = param%d;\n", i + 1);
		}
		emit("    *gen_opc_ptr++ = INDEX_%s;\n", fi->sym->name);
		emit("}\n\n");
	}
}

void
emit_translator_case_pushrelocs(vlong mos)
{
	Rrec *ri=specialop_push->rrecs;

	emit("\t\t\tdyngen_itab[%d]->addr = (ulong)(gen_code_ptr + %lld);\n",
		ri->reloc->reloc_index, specialop_push->size + mos);

	emit("\t\t\tdynreloc(gen_code_ptr - %ulld,"
			" %uld, %d, dyngen_itab, dyngen_nimport);\n",
			specialop_push->bounds[0],
			ri->reloc->offset,
			ri->reloc->mode);
}


void
emit_translator_case_relocs(Funct *fi)
{
	Rrec *ri;
	int i;

	for(i = 1; i <= fi->maxcfold ; i++)
	{
		emit("\t\t\tulong param%d = *opparam_ptr++;\n", i);
	}

	for(ri=fi->rrecs; ri != nil; ri=ri->next){
		int cfold = import_table_meta[ri->reloc->reloc_index].cfold_index;
		int label = import_table_meta[ri->reloc->reloc_index].label_index;
		if(cfold != -1){
			emit("\t\t\tdyngen_itab[%d]->addr = param%d;\n",
				ri->reloc->reloc_index, cfold);
		}else if(label != -1){
			emit("\t\t\tdyngen_itab[%d]->addr = gen_labels[param%d];\n",
				ri->reloc->reloc_index, label);
		}

		emit("\t\t\tdynreloc(gen_code_ptr - %ulld,"
				" %uld, %d, dyngen_itab, dyngen_nimport);",
				fi->bounds[0],
				ri->reloc->offset,
				ri->reloc->mode);

		if(verbose)
			emit("\t/* ri=%d, ro=0x%llx */\n",
				ri->reloc->reloc_index,
				ri->reloc->reloc_offset);
		else
			emit("\n");

	}
}

void
emit_translator_case_push(Funct *fi)
{
	emit("\t\t{\n");
	/* Everybody's favorite special push operation */
	emit("\t\t\textern uchar %s[];\n", specialop_push->sym->name);
	emit("\t\t\tmemcpy(gen_code_ptr, %s, %ulld);\n",
		specialop_push->sym->name,
		specialop_push->size
	);

	emit_translator_case_pushrelocs(fi->size);

	emit("\t\t\tgen_code_ptr += %ulld;\n", specialop_push->size);

	emit("\t\t}\n");
}

void
emit_translator_case_tail(Funct *fi)
{
	emit("\t\t{\n");

	/* The actual micro-op */
	emit("\t\t\textern uchar %s[];\n", fi->sym->name);
	emit("\t\t\tmemcpy(gen_code_ptr, %s, %ulld);\n",
		fi->sym->name,
		fi->size
	);

	emit_translator_case_relocs(fi);

	emit("\t\t\tgen_code_ptr += %ulld;\n", fi->size);

	emit("\t\t}\n\t\tbreak;\n");
}


void
emit_translator_case(Funct *fi)
{
	emit_translator_case_push(fi);
	emit_translator_case_tail(fi);
}

void
emit_translator(void){
	Funct *fi;

	emit("extern Dynsym *dyngen_itab[];\n");
	emit("extern int dyngen_nimport;\n");

	emit(
		"int dyngen_code(uint8_t *gen_code_buf,\n"
		"				uint16_t *label_offsets, uint16_t *jmp_offsets,\n"
		"				const uint16_t *opc_buf, const uint32_t *opparam_buf,\n"
		"				const long *gen_labels)\n"
		"{\n"
		"	uchar *gen_code_ptr;\n"
		"	const u16int *opc_ptr;\n"
		"	const u32int *opparam_ptr;\n"
	);

	emit(
		"	gen_code_ptr = gen_code_buf;\n"
		"	opc_ptr = opc_buf;\n"
		"	opparam_ptr = opparam_buf;\n"
	);

	emit(
		"	for(;;) {\n"
	);

	emit(
		"		switch(*opc_ptr++) {\n"
	);


	for(fi=microops; fi != nil; fi=fi->next){
		if(fi->qemu_op_name == nil)
			continue;

		emit("\t\tcase INDEX_%s%s:\n", OP_PREFIX, fi->qemu_op_name);
		emit_translator_case(fi);
	}

	/* Plan 9 special operations */
	emit("\t\tcase INDEX_%sp9_storestack:\n", OP_PREFIX);
	emit_translator_case_tail(specialop_storestack);

	/* NO-OP nonsense and default */
	emit("\t\tcase INDEX_op_nop:\n\t\tbreak;\n");
    emit("\t\tcase INDEX_op_nop1:\n\t\t\topparam_ptr++;\n\t\tbreak;\n");
    emit("\t\tcase INDEX_op_nop2:\n\t\t\topparam_ptr+=2;\n\t\tbreak;\n");
    emit("\t\tcase INDEX_op_nop3:\n\t\t\topparam_ptr+=3;\n\t\tbreak;\n");
	emit("\t\tdefault:\n\t\t\tgoto the_end;\n");
	emit("\t\t}\n\t}\n");

	emit("the_end:\n");
	emit("\tflush_icache_range(gen_code_buf, gen_code_ptr);\n");
	emit("\treturn gen_code_ptr -  gen_code_buf;\n");
	emit("}\n");
}

void
usage(void)
{
	print("%s: -{c|g|t|i|b} [-p] [-f opsfile] [-o outfile]\n", argv0);
	print("Modes:\n"
		"b: emit micro-op bodies as C character arrays.\n"
		"c: emit DEF()s for micro-ops.  Name is from Dyngen/UNIX.\n"
		"g: emit micro-op generator C functions.\n"
		"i: emit the import table.\n"
		"t: emit the translator main function, dyngen_code.\n"
	);
	print("Misc:\n"
		"f opsfile: specify the dlm containing the micro-ops\n"
		"o outfile: write output to outfile; default is /fd/1.\n"
		"p: Increase paranoia.\n"
		"v: Increase verbosity.\n"
	);

	exits("usage");
}

void
main(int argc, char **argv)
{
	char *inpath = "./opsdlm";
	char *outpath = "/fd/1";
	int outfd;
	mode = MODE_USAGE;


	ARGBEGIN{
		/* Mode toggles */
	case 'b':
		mode = MODE_OP_BODIES;
		break;
	case 'c':
		mode = MODE_OP_DEFINES;
		break;
	case 'g':
		mode = MODE_OP_GENERATORS;
		break;
	case 'i':
		mode = MODE_TABLES;
		break;
	case 't':
		mode = MODE_TRANSLATOR;
		break;

		/* Input/Output */
	case 'f':
		inpath = EARGF(usage());
		break;
	case 'o':
		outpath = EARGF(usage());
		break;

		/* Misc */
	case 'p':
		paranoia++;
		break;
	case 'v':
		verbosity++;
		break;

	default:
		usage();
		return;
	}ARGEND

	if(mode == MODE_USAGE){
		usage();
	}

	if(verbosity > paranoia)
		paranoid = verbose;
	else if(paranoia == 0)
		paranoid = nullprint;

	outfd = create(outpath, OWRITE | OTRUNC, 0744);
	if(outfd < 0)
		fatal("Could not open output file '%s': %r\n", outpath);
	Binit(&outbio, outfd, OWRITE);

	infd = open(inpath, OREAD);
	if(infd < 0)
		fatal("cannot open input file '%s': %r", inpath);

	seek(infd, 0, 0);
	if (crackhdr(infd, &fh) == 0) {
		fatal("Can't read header: %r\n");
	}

	if (syminit(infd, &fh) < 0)
		fatal("Syminit: %r\n");

	switch(mode){
		case MODE_OP_BODIES:
			filter_symbols();
			load_text();
			emit_functions();
			exits(nil);
		case MODE_TABLES:
			load_imports();
			create_import_meta();
			emit_import_table();
			exits(nil);
	}

	filter_symbols();

	load_imports();
	create_import_meta();

	if(verbosity)
		verbose_import_table();

	load_relocs();
	pair_relocs();

	if(paranoia)
		check_cfolds();

	if(mode == MODE_OP_GENERATORS){
		emit_generators();
		exits(nil);
	}

	if(paranoia)
		check_special_ops();


	switch(mode){
		case MODE_OP_DEFINES:
		{
			uvlong size = 0;

			if(specialop_push != nil)
				size = specialop_push->size;

			verbose("/* Growing all micro-op bodies by %ulld */\n", size);

			emit_definitions(size);
			break;
		}
		case MODE_TRANSLATOR:
			emit_translator();
			break;
	}

	exits(nil);
}
